#! python3  # noqa: E265

"""
    Main command-line.
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from os import environ
from pathlib import Path
from timeit import default_timer

# 3rd party library
import click

# submodules
from qgis_deployment_toolbelt.__about__ import __version__
from qgis_deployment_toolbelt.commands import cli_check, cli_clean
from qgis_deployment_toolbelt.jobs import JobsOrchestrator
from qgis_deployment_toolbelt.scenarios import ScenarioReader
from qgis_deployment_toolbelt.utils.bouncer import exit_cli_error, exit_cli_normal

# #############################################################################
# ########## Globals ###############
# ##################################

# chronometer
START_TIME = default_timer()

# logs
logger = logging.getLogger(__name__)
log_console_handler = logging.StreamHandler()
logger.addHandler(log_console_handler)

# default CLI context.
# See: https://click.palletsprojects.com/en/7.x/commands/#context-defaults
CONTEXT_SETTINGS = dict(obj={})

# #############################################################################
# ####### Command-line ############
# #################################


@click.group(
    chain=True,
    invoke_without_command=True,
    context_settings=CONTEXT_SETTINGS,
)
@click.option(
    "-c",
    "--clear",
    is_flag=True,
    show_default=True,
    help="Clear the terminal before the execution.",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    show_default=True,
    help="Set output verbosity to the maximum level, overriding the configuration option.",
)
@click.option(
    "-s",
    "--scenario",
    "scenario_filepath",
    default="scenario.qdt.yml",
    show_default=True,
    help="Scenario file to use.",
    type=click.Path(readable=True, file_okay=True, dir_okay=False, resolve_path=True),
)
@click.version_option(
    version=__version__,
    message="%(version)s",
    help="Display CLI version",
)
@click.pass_context
def qgis_deployment_toolbelt(
    cli_context: click.Context,
    scenario_filepath: Path,
    clear: bool,
    verbose: bool,
):
    """Main command.

    \f
    Args:
        cli_context (click.Context): Click context
        scenario_filepath (Path): path to a scenario file to use
        clear (bool): option to clear the terminal berfore any other step
        verbose (bool): option to force the verbose mode

    :Example:

        .. code-block:: powershell

            qgis-deployment-toolbelt -c --verbose check

    """
    # let's be clear or not
    if clear:
        click.clear()

    # -- LOG/VERBOSITY MANAGEMENT ------------------------------------------------------
    # if verbose, override conf value
    if verbose:
        logger.setLevel(logging.DEBUG)
        for h in logger.handlers:
            h.setLevel(logging.DEBUG)
    logger.info(f"{logging.getLevelName(logger.getEffectiveLevel())} mode enabled.")

    click.echo(
        "Timestamp: {} started after {:5.2f}s.".format(
            cli_context.info_name, default_timer() - START_TIME
        )
    )

    # -- USING DEFAULT SCENARIO OR NOT -------------------------------------------------
    if cli_context.invoked_subcommand is None and Path(scenario_filepath).is_file():
        logger.debug(
            f"Straight run launched using default scenario file: {scenario_filepath}."
        )

        # -- LOAD CONFIGURATION FILE ---------------------------------------------------
        scenario = ScenarioReader(in_yaml=scenario_filepath)

        # Apply log level from scenario (only if verbose mode is disabled)
        if scenario.settings.get("DEBUG") is True and not verbose:
            logger.setLevel(logging.DEBUG)
            for h in logger.handlers:
                h.setLevel(logging.DEBUG)
            logger.info(
                f"{logging.getLevelName(logger.getEffectiveLevel())} mode enabled."
            )

        # Validate scenario
        if (
            scenario.settings.get("SCENARIO_VALIDATION", True) is True
            and not scenario.validate_scenario()[0]
        ):
            exit_cli_error(
                "Scenario validation failed. Please check the scenario file."
                "\nValidation report:\n- {}".format(
                    "\n- ".join(scenario.validate_scenario()[1])
                )
            )

        # Use metadata to inform which scenario is running
        click.echo(
            "Running scenario: {title} ({id}).\n{description}".format(
                **scenario.metadata
            )
        )

        # Set environment vars for the scenario
        for var, value in scenario.settings.items():
            if value is not None:
                logger.debug(f"Setting environment variable {var} = {value}.")
                environ[var] = str(value)
            else:
                logger.debug(f"Ignored None value: {var}.")

        # -- STEPS JOBS
        steps_ok = []
        orchestrator = JobsOrchestrator()

        # filter out unrecognized jobs
        for step in scenario.steps:
            if step.get("uses") not in orchestrator.jobs_ids:
                logger.warning(
                    f"{step.get('uses')} not found in available jobs. Skipping."
                )
                continue
            else:
                steps_ok.append(step)

        # run job
        with click.progressbar(
            steps_ok, label="Running the scenario.."
        ) as progress_bar:
            for step in progress_bar:
                click.secho("Running step: {}".format(step.get("uses")), fg="green")
                try:
                    job = orchestrator.init_job_class_from_id(
                        job_id=step.get("uses"), options=step.get("with")
                    )
                    job.run()
                except Exception as err:
                    exit_cli_error(err)

    # -- ERROR -------------------------------------------------------------------------
    elif cli_context.invoked_subcommand is None and not Path(scenario).is_file():
        exit_cli_error(
            "Straight run launched but no default scenario file found."
            "\nPlease make sure there is a default scenario file `scenario.qdt.yml` "
            f"here {Path(scenario).parent} or use it as a CLI passing the scenario "
            "filepath as an argument."
        )
    else:
        logger.debug(
            f"CLI mode enabled, invoking {cli_context.invoked_subcommand} "
            f"with arguments: {cli_context.args}."
        )
        exit_cli_normal(message="CLI mode enabled", abort=False)

    # end
    logger.debug(
        "Timestamp: {} completed after {:5.2f}s.".format(
            cli_context.info_name, default_timer() - START_TIME
        )
    )


# -- SUB-COMMANDS ----------------------------------------------------------------------
# Add subcommands to the main command group
qgis_deployment_toolbelt.add_command(cli_check.check)
qgis_deployment_toolbelt.add_command(cli_clean.clean)

# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    # launch cli
    qgis_deployment_toolbelt(obj={})



#! python3  # noqa: E265

"""
    Utilities specific for Windows.

    Author: Julien Moura (https://github.com/guts)
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from sys import platform as opersys
from typing import Union

# Imports depending on operating system
if opersys == "win32":
    """windows"""
    import winreg
else:
    pass

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Functions #############
# ##################################
def get_environment_variable(envvar_name: str, scope: str = "user") -> Union[str, None]:
    """Get environment variable from Windows registry.

    :param str envvar_name: environment variable name to retrieve
    :param str scope: . Must be "user" or "system", defaults to "user".

    :return Union[str, None]: environment variable value or None if not found
    """
    # user or system
    if scope == "user":
        hkey = (winreg.HKEY_CURRENT_USER, r"Environment")
    else:
        hkey = (
            winreg.HKEY_LOCAL_MACHINE,
            r"SYSTEM\CurrentControlSet\Control\Session Manager\Environment",
        )
    # try to get the value
    try:
        with winreg.OpenKey(*hkey, access=winreg.KEY_READ) as key:
            value, regtype = winreg.QueryValueEx(key, envvar_name)
        return value
    except OSError:
        logger.error(
            f"Environment variable {envvar_name} not found in registry (scope: {scope}"
        )
        return None


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265

"""
    Metadata bout the package to easily retrieve informations about it.
    See: https://packaging.python.org/guides/single-sourcing-package-version/
"""

from datetime import date

__all__ = [
    "__author__",
    "__copyright__",
    "__email__",
    "__license__",
    "__summary__",
    "__title__",
    "__uri__",
    "__version__",
]

__author__ = "Julien Moura (Oslandia)"
__copyright__ = "2021 - {0}, {1}".format(date.today().year, __author__)
__email__ = "qgis@oslandia.com"
__executable_name__ = "QGIS_DeploymentCLI.exe"
__package_name__ = "qgis-deployment-toolbelt"
__keywords__ = ["cli, QGIS, deployment, profiles"]
__license__ = "GNU Lesser General Public License v3.0"
__summary__ = (
    "QGIS deployment toolbelt is a CLI (Command Line Interface) to perform "
    "redundant operations after a QGIS deployment."
)
__title__ = "QGIS Deployment Toolbelt"
__title_clean__ = "".join(e for e in __title__ if e.isalnum())
__uri_homepage__ = "https://guts.github.io/qgis-deployment-cli/"
__uri_repository__ = "https://github.com/Guts/qgis-deployment-cli/"
__uri_tracker__ = "https://github.com/Guts/qgis-deployment-cli/issues/"
__uri__ = __uri_repository__

__version__ = "0.9.0"
__version_info__ = tuple(
    [
        int(num) if num.isdigit() else num
        for num in __version__.replace("-", ".", 1).split(".")
    ]
)



#! python3  # noqa: E265

"""
    Log management.

    Author: Julien Moura (github.com/guts)

    See: https://docs.python.org/fr/3/howto/logging.html
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from getpass import getuser
from logging.handlers import RotatingFileHandler
from os import environ
from pathlib import Path
from platform import architecture
from platform import platform as opersys
from socket import gethostname

# modules
from qgis_deployment_toolbelt.__about__ import __title__
from qgis_deployment_toolbelt.__about__ import __title_clean__ as package_name
from qgis_deployment_toolbelt.__about__ import __version__
from qgis_deployment_toolbelt.utils.bouncer import exit_cli_error

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class LogManager:
    """Manage logs: configuration, parsing, etc."""

    LOG_FORMAT = logging.Formatter(
        "%(asctime)s || %(levelname)s "
        "|| %(module)s - %(lineno)d ||"
        " %(funcName)s || %(message)s"
    )

    def __init__(
        self,
        debug_level: int = 0,
        label: str = package_name,
        folder: Path = Path("./_logs"),
    ):
        """Instanciation method."""
        self.debug_level = debug_level
        self.folder = Path(folder)
        self.label = "".join(e for e in label if e.isalnum())

        if debug_level == 2:
            self.console_level = logging.DEBUG
            self.file_level = logging.DEBUG
            self.log_level = "DEBUG"
        elif debug_level == 1:
            self.console_level = logging.INFO
            self.file_level = logging.INFO
            self.log_level = "INFORMATIVE"
        else:
            self.console_level = logging.WARNING
            self.file_level = logging.INFO
            self.log_level = "LIGHT"

        # ensure folder is created
        try:
            folder.mkdir(exist_ok=True, parents=True)
        except PermissionError as err:
            msg_err = (
                "Impossible to create the logs folder. Does the user '{}' ({}) have "
                "write permissions on: {}. Trace: {}"
            ).format(environ.get("userdomain"), getuser(), folder, err)
            exit_cli_error(msg_err)
        self.folder = folder

        # create logger
        self.initial_logger_config()

    def initial_logger_config(self) -> logging.Logger:
        """Configure root logger. \
        BE CAREFUL: it depends a lot of how click implemented logging facilities. \
        So, sadly, every option is not available.

        :return: configured logger
        :rtype: logging.Logger
        """
        #  create main logger
        logging.captureWarnings(False)
        logger = logging.getLogger()
        logger.setLevel(self.console_level)

        # create console handler - seems to be ignored by click
        log_console_handler = logging.StreamHandler()
        log_console_handler.setLevel(self.console_level)

        # create file handler
        log_file_handler = RotatingFileHandler(
            filename=self.folder / "{}.log".format(self.label),
            mode="a",
            maxBytes=3000000,
            backupCount=10,
            encoding="UTF-8",
        )
        log_file_handler.setLevel(self.file_level)

        # apply format
        log_file_handler.setFormatter(self.LOG_FORMAT)

        # add only file handler to the logger, to avoid duplicated messages
        # logger.addHandler(log_console_handler)
        logger.addHandler(log_file_handler)

        return logger

    def headers(self):
        """Basic information to log before other message."""
        # initialize the log
        logger.info("===== {} - Version {} =====".format(__title__, __version__))
        logger.info("Operating System: {}".format(opersys()))
        logger.info("Architecture: {}".format(architecture()[0]))
        logger.info("Computer: {}".format(gethostname()))
        logger.info("Launched by: {}".format(getuser()))
        logger.info("OS Domain: {}".format(environ.get("userdomain")))


# #############################################################################
# ##### Main #######################
# ##################################
if __name__ == "__main__":
    pass



#! python3  # noqa: E265

"""
    Read and validate scenario files.

    Author: Julien Moura (https://github.com/guts, Oslandia)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from functools import lru_cache
from io import BufferedIOBase
from os import R_OK, access
from pathlib import Path
from typing import List, Tuple, Union

# 3rd party
import yaml

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class ScenarioReader:
    """Read and validate scenario files."""

    scenario: dict = None

    def __init__(self, in_yaml: Union[str, Path, BufferedIOBase]):
        """Instanciating YAML scenario reader."""
        # check and get YAML path
        if isinstance(in_yaml, (str, Path)):
            self.input_yaml = self.check_yaml_file(in_yaml)
            # extract data from input file
            with self.input_yaml.open(mode="r") as bytes_data:
                self.scenario = yaml.safe_load(bytes_data)
        elif isinstance(in_yaml, BufferedIOBase):
            self.input_yaml = self.check_yaml_buffer(in_yaml)
            # extract data from input file
            self.scenario = yaml.safe_load(self.input_yaml)
        else:
            raise TypeError

    # CHECKS
    def check_yaml_file(self, yaml_path: Union[str, Path]) -> Path:
        """Perform some checks on passed yaml file and load it as Path object.

        :param yaml_path: path to the yaml file to check

        :returns: sanitized yaml path
        :rtype: Path
        """
        # if path as string load it in Path object
        if isinstance(yaml_path, str):
            try:
                yaml_path = Path(yaml_path)
            except Exception as exc:
                raise TypeError("Converting yaml path failed: {}".format(exc))

        # check if file exists
        if not yaml_path.exists():
            raise FileExistsError(
                "YAML file to check doesn't exist: {}".format(yaml_path.resolve())
            )

        # check if it's a file
        if not yaml_path.is_file():
            raise IOError("YAML file is not a file: {}".format(yaml_path.resolve()))

        # check if file is readable
        if not access(yaml_path, R_OK):
            raise IOError("yaml file isn't readable: {}".format(yaml_path))

        # check integrity and structure
        with yaml_path.open(mode="r") as in_yaml_file:
            try:
                yaml.safe_load_all(in_yaml_file)
            except yaml.YAMLError as exc:
                logger.error(msg="YAML file is invalid: {}".format(yaml_path.resolve()))
                raise exc
            except Exception as exc:
                logger.error(msg="Structure of YAML file is incorrect: {}".format(exc))
                raise exc

        # return sanitized path
        return yaml_path

    def check_yaml_buffer(self, yaml_buffer: BufferedIOBase) -> BufferedIOBase:
        """Perform some checks on passed yaml file.

        :param yaml_buffer: bytes reader of the yaml file to check

        :returns: checked bytes object
        :rtype: BufferedIOBase
        """
        # check integrity
        try:
            yaml.safe_load_all(yaml_buffer)
        except yaml.YAMLError as exc:
            logger.error("Invalid YAML {}. Trace: {}".format(yaml_buffer, exc))
            raise exc

        # return sanitized path
        return yaml_buffer

    @lru_cache
    def validate_scenario(self) -> Tuple[bool, List[str]]:
        """Validate scenario file.

        TODO: use json schema to validate scenario file.

        :returns: True if scenario is valid, False otherwise and a report of validation
        errors (which is None if the scenario is valid).
        :rtype: Tuple[bool, List[str]]
        """
        # variables
        required_root_keys: tuple = ("metadata", "settings", "steps")

        # outputs
        valid: bool = True
        report: List[str] = []

        # check if scenario is a dict
        if not isinstance(self.scenario, dict):
            report.append(f"Scenario is not a dict but {type(self.scenario)}")
            valid = False

        # check scenario basic structure
        if any([i not in self.scenario for i in required_root_keys]):
            report.append(
                "Some of required root keys are missing: {}".format(
                    ", ".join(required_root_keys)
                )
            )
            valid = False

        # check if metadata is a dict
        if not isinstance(self.metadata, dict):
            report.append("Metadata is not a dict: {}".format(self.metadata))
            valid = False

        return valid, report

    @property
    def metadata(self) -> dict:
        """Get metadata from scenario.

        :returns: metadata
        :rtype: dict
        """
        return self.scenario.get("metadata")

    @property
    def settings(self) -> dict:
        """Get toolbelt settings from scenario.

        :returns: settings
        :rtype: dict
        """
        return self.scenario.get("settings")

    @property
    def steps(self) -> List[dict]:
        """Get steps from scenario.

        :returns: steps
        :rtype: List[dict]
        """
        return self.scenario.get("steps")


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    from pprint import pprint

    reader = ScenarioReader(
        Path("tests/fixtures/scenarios/good_scenario_sample.qdt.yml")
    )
    pprint(reader.scenario)



#! python3  # noqa: E265

"""
    Tools to manage the environment setup (variables, etc.)

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from os.path import expanduser, expandvars
from pathlib import Path
from sys import platform as opersys
from typing import List

# Imports depending on operating system
if opersys == "win32":
    """windows"""
    import win32gui
    from py_setenv import setenv
else:
    pass

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class JobEnvironmentVariables:
    """
    Class to manage the environment variables of QGIS installation.
    """

    ID: str = "manage-env-vars"

    def __init__(self, options: List[dict]) -> None:
        """Instantiate the class.

        :param List[dict] options: dictionary with environment variable name as key and
        some parameters as values (value, scope, action...).
        """

        self.options: dict = self.validate_options(options)

    def run(self) -> None:
        """Apply environment variables from dictionary to the system."""
        if opersys == "win32":
            for env_var in self.options:
                if env_var.get("action") == "add":
                    setenv(
                        name=env_var.get("name"),
                        value=self.prepare_value(env_var.get("value")),
                        user=env_var.get("scope") == "user",
                        suppress_echo=True,
                    )
            # force Windows to refresh the environment
            self.win_refresh_environment()

        # TODO: for linux, edit ~/.profile or add a .env file and source it from ~./profile
        else:
            logger.debug(
                f"Setting persistent environment variables is not supported on {opersys}"
            )

        logger.debug(f"Job {self.ID} ran successfully.")

    def prepare_value(self, value: str) -> str:
        """Prepare value to be used in the environment variable.

        :param str value: value to prepare.
        :return str: prepared value.
        """
        try:
            # test if value is a path
            value_as_path = Path(expanduser(expandvars(value)))
            if not value_as_path.exists():
                logger.warning(
                    f"{value} seems to be a valid path but does not exist (yet)."
                )

            return str(value_as_path.resolve())
        except Exception as err:
            logger.debug(f"Value {value} is not a valid path: {err}")

        if opersys == "win32":
            return value
        else:
            return f'"{value}"'

    def validate_options(self, options: List[dict]) -> List[dict]:
        """Validate options.

        :param List[dict] options: options to validate.
        :return List[dict]: options if they are valid.
        """
        if not isinstance(options, list):
            raise TypeError(f"Options must be a list, not {type(options)}")
        for option in options:
            if not isinstance(option, dict):
                raise TypeError(f"Options must be a dict, not {type(option)}")

        return options

    def win_refresh_environment(self) -> bool:
        """This ensures that changes to Windows registry are immediately propagated.
        Useful to refresh after have updated the environment variables.

        A method by Geoffrey Faivre-Malloy and Ronny Lipshitz.
        Source: https://gist.github.com/apetrone/5937002

        :return bool: True if the environment has been refreshed
        """
        # broadcast settings change
        HWND_BROADCAST: int = 0xFFFF
        WM_SETTINGCHANGE: int = 0x001A
        SMTO_ABORTIFHUNG: int = 0x0002
        sParam = "Environment"

        res1, res2 = win32gui.SendMessageTimeout(
            HWND_BROADCAST, WM_SETTINGCHANGE, 0, sParam, SMTO_ABORTIFHUNG, 100
        )
        if not res1:
            logger.warning(
                f"Refresh environment failed: {bool(res1)}, {res2}, from SendMessageTimeout"
            )
            return False
        else:
            return True


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265

"""
    Download remote QGIS profiles to local.

    Author: Julien Moura (https://github.com/guts)
"""


# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from os.path import expanduser, expandvars
from pathlib import Path
from shutil import copy2, copytree
from sys import platform as opersys
from typing import Tuple

# package
from qgis_deployment_toolbelt.constants import OS_CONFIG
from qgis_deployment_toolbelt.profiles import RemoteGitHandler

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################


class JobProfilesDownloader:
    """
    Job to download remote profiles and set them.
    """

    ID: str = "qprofiles-manager"
    OPTIONS_SCHEMA: dict = {
        "action": {
            "type": str,
            "required": False,
            "default": "download",
            "possible_values": ("download", "refresh"),
            "condition": "in",
        },
        "local_destination": {
            "type": str,
            "required": False,
            "default": ".cache/qgis-deployment-toolbelt/profiles",
            "possible_values": None,
            "condition": None,
        },
        "protocol": {
            "type": str,
            "required": True,
            "default": "http",
            "possible_values": ("http", "git", "copy"),
            "condition": "in",
        },
        "source": {
            "type": str,
            "required": True,
            "default": None,
            "possible_values": ("https://", "http://", "git://", "file://"),
            "condition": "startswith",
        },
    }
    PROFILES_NAMES_DOWNLOADED: list = []
    PROFILES_NAMES_INSTALLED: list = []

    def __init__(self, options: dict) -> None:
        """Instantiate the class.

        :param dict options: profiles source (remote, can be a local network) and
        destination (local).
        """
        self.options: dict = self.validate_options(options)

        # profile folder
        if opersys not in OS_CONFIG:
            raise OSError(
                f"Your operating system {opersys} is not supported. "
                f"Supported platforms: {','.join(OS_CONFIG.keys())}."
            )
        self.qgis_profiles_path: Path = Path(OS_CONFIG.get(opersys).profiles_path)
        if not self.qgis_profiles_path.exists():
            logger.warning(
                f"QGIS profiles folder not found: {self.qgis_profiles_path}. "
                "Creating it to properly run the job."
            )
            self.qgis_profiles_path.mkdir(parents=True)

        # TODO: handle custom profiles folder through QGIS_CUSTOM_CONFIG_PATH

        # list installed profiles
        self.PROFILES_NAMES_INSTALLED = [
            d.name
            for d in self.qgis_profiles_path.iterdir()
            if d.is_dir() and not d.name.startswith(".")
        ]

        # prepare local destination
        self.local_path: Path = Path(
            expandvars(expanduser(self.options.get("local_destination")))
        )
        if not self.local_path.exists():
            self.local_path.mkdir(parents=True, exist_ok=True)

    def run(self) -> None:
        """Execute job logic."""
        # download or refresh
        if self.options.get("action") != "download":
            raise NotImplementedError

        # prepare remote source
        if self.options.get("protocol") == "git":
            downloader = RemoteGitHandler(url=self.options.get("source"))
            downloader.clone_or_pull(self.local_path)
        else:
            raise NotImplementedError

        # check of there are some profiles folders within the downloaded folder
        profiles_folders = self.filter_profiles_folder()
        if profiles_folders is None:
            logger.error("No QGIS profile found in the downloaded folder.")
            return

        # store downloaded profiles names
        self.PROFILES_NAMES_DOWNLOADED = [d.name for d in profiles_folders]
        logger.info(
            f"{len(self.PROFILES_NAMES_DOWNLOADED)} downloaded profiles: "
            f"{', '.join(self.PROFILES_NAMES_DOWNLOADED)}"
        )

        # copy profiles to the QGIS 3
        self.sync_local_profiles(source_profiles_folder=profiles_folders)

        logger.debug(f"Job {self.ID} ran successfully.")

    def filter_profiles_folder(self) -> Tuple[Path] or None:
        """Parse downloaded folder to filter on QGIS profiles folders.

        :return Tuple[Path] or None: tuple of profiles folders paths
        """
        # first, try to get folders containing a profile.json
        qgis_profiles_folder = [
            f.parent for f in self.local_path.glob("**/profile.json")
        ]
        if len(qgis_profiles_folder):
            return tuple(qgis_profiles_folder)

        # if empty, try to identify if a folder is a QGIS profile - but unsure
        for d in self.local_path.glob("**"):
            if (
                d.is_dir()
                and d.parent.name == "profiles"
                and not d.name.startswith(".")
            ):
                qgis_profiles_folder.append(d)

        if len(qgis_profiles_folder):
            return tuple(qgis_profiles_folder)

        # if still empty, raise a warning but returns every folder under a `profiles` folder
        # TODO: try to identify if a folder is a QGIS profile with some approximate criteria

        if len(qgis_profiles_folder):
            logger.error("No QGIS profile found in the downloaded folder.")
            return None

    def sync_local_profiles(self, source_profiles_folder: tuple) -> None:
        """Copy downloaded profiles to QGIS profiles folder.
        If the QGIS profiles folder doesn't exist, it will be created and every
        downloaded profile will be copied.
        If a profile is already installed, it won't be overwritten.

        :param tuple source_profiles_folder: list of downloaded profiles folders paths
        """
        # check if local profiles folder exists or it's empty
        if not self.qgis_profiles_path.exists() or not any(
            self.qgis_profiles_path.iterdir()
        ):
            # ensure it exists
            self.qgis_profiles_path.mkdir(parents=True, exist_ok=True)
            # copy downloaded profiles into this
            for d in source_profiles_folder:
                copytree(
                    d,
                    self.qgis_profiles_path,
                    copy_function=copy2,
                    dirs_exist_ok=True,
                )
        elif len(
            set(self.PROFILES_NAMES_DOWNLOADED) - set(self.PROFILES_NAMES_INSTALLED)
        ):
            already_installed = [
                p
                for p in self.PROFILES_NAMES_DOWNLOADED
                if p in self.PROFILES_NAMES_INSTALLED
            ]
            not_installed = [
                p
                for p in self.PROFILES_NAMES_DOWNLOADED
                if p not in self.PROFILES_NAMES_INSTALLED
            ]

            logger.error(
                "Mixed case. "
                f"Already installed profiles: {','.join(already_installed)}. "
                f"Not installed: {','.join(not_installed)}."
            )

            for d in source_profiles_folder:
                if d.name in not_installed:
                    # create destination parent folder
                    to_profile_parent_folderpath = Path(
                        self.qgis_profiles_path / d.name
                    )
                    to_profile_parent_folderpath.mkdir(parents=True, exist_ok=True)
                    copytree(
                        d,
                        to_profile_parent_folderpath,
                        copy_function=copy2,
                        dirs_exist_ok=True,
                    )

        else:
            logger.error(
                "QGIS Profiles folder already exists, it's not empty: "
                f"{self.qgis_profiles_path.resolve()}"
            )

    def validate_options(self, options: dict) -> bool:
        """Validate options.

        :param dict options: options to validate.
        :return bool: True if options are valid.
        """
        for option in options:
            if option not in self.OPTIONS_SCHEMA:
                raise Exception(
                    f"Job: {self.ID}. Option '{option}' is not valid."
                    f" Valid options are: {self.OPTIONS_SCHEMA.keys()}"
                )

            option_in = options.get(option)
            option_def: dict = self.OPTIONS_SCHEMA.get(option)
            # check value type
            if not isinstance(option_in, option_def.get("type")):
                raise Exception(
                    f"Job: {self.ID}. Option '{option}' has an invalid value."
                    f"\nExpected {option_def.get('type')}, got {type(option_in)}"
                )
            # check value condition
            if option_def.get("condition") == "startswith" and not option_in.startswith(
                option_def.get("possible_values")
            ):
                raise Exception(
                    f"Job: {self.ID}. Option '{option}' has an invalid value."
                    "\nExpected: starts with one of: "
                    f"{', '.join(option_def.get('possible_values'))}"
                )
            elif option_def.get(
                "condition"
            ) == "in" and option_in not in option_def.get("possible_values"):
                raise Exception(
                    f"Job: {self.ID}. Option '{option}' has an invalid value."
                    f"\nExpected: one of: {', '.join(option_def.get('possible_values'))}"
                )
            else:
                pass

        return options


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265

"""
    Basic string to bool.
"""

# #############################################################################
# ########## Globals ###############
# ##################################

_true_set = {"on", "t", "true", "y", "yes", "1"}
_false_set = {"f", "false", "n", "no", "off", "0"}


# #############################################################################
# ########## Functions #############
# ##################################
def str2bool(value: str, raise_exc: bool = False) -> bool:
    """Determine if a string is a bool and, if so, convert it.

    :param str value: value to convert
    :param bool raise_exc: strict mode, defaults to False
    :raises ValueError: value is not in true and false sets

    :return bool: True if input string is in _true_set, False if it's in _false_set.
    None or Exception if not in any of two sets.
    """
    if isinstance(value, str):
        value = value.lower()
        if value in _true_set:
            return True
        if value in _false_set:
            return False

        if raise_exc:
            raise ValueError('Expected "%s"' % '", "'.join(_true_set | _false_set))

        return None
    else:
        raise TypeError(f"value must be a str, not {type(value)}")



#! python3  # noqa: E265

"""Setup script to package into a Python module."""

# ############################################################################
# ########## Libraries #############
# ##################################

# standard library
from pathlib import Path

# 3rd party
from setuptools import find_packages, setup

# package (to get version)
from qgis_deployment_toolbelt import __about__

# ############################################################################
# ########### Globals ##############
# ##################################

# The directory containing this file
HERE = Path(__file__).parent

with open(HERE / "requirements/base.txt") as f:
    requirements = f.read().splitlines()

# The text of the README file
README = (HERE / "README.md").read_text()

# ############################################################################
# ########### Main #################
# ##################################

setup(
    name=__about__.__package_name__,
    author=__about__.__author__,
    author_email=__about__.__email__,
    description=__about__.__summary__,
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=__about__.__keywords__,
    url=__about__.__uri__,
    version=__about__.__version__,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: GIS",
        "Topic :: System :: Installation/Setup",
    ],
    # packaging
    py_modules=["qgis_deployment_toolbelt"],
    packages=find_packages(
        exclude=["contrib", "docs", "*.tests", "*.tests.*", "tests.*", "tests", ".venv"]
    ),
    include_package_data=True,
    install_requires=requirements,
    entry_points="""
        [console_scripts]
        qdeploy-toolbelt=qgis_deployment_toolbelt.cli:qgis_deployment_toolbelt
        qgis-deployment-toolbelt=qgis_deployment_toolbelt.cli:qgis_deployment_toolbelt
    """,
)
